import numpy as np

def pca_projection(X, k=2):
    Xn, mu, sigma = normalize_features(X)
    U, S, V = pca(Xn)
    return project_data(X, U, k)


def project_data(X, U, K):
    Z = np.zeros((X.shape[0], K))
    for i in range(0, X.shape[0]):
        for k in range(0, K):
            x = X[i, :]
            Z[i, k] = x.dot(U[:, k])
    return Z


def pca(X):
    m = X.shape[0]
    the_sigma = X.transpose().dot(X) / m
    return np.linalg.svd(the_sigma)


def normalize_features(X):
    mu = np.mean(X, 0)
    Xn = X - mu
    sigma = np.std(Xn, 0)
    for i in range(sigma.shape[0]):
        if sigma[i] == 0:
            sigma[i] = 1
    Xn = Xn / sigma
    return Xn, mu, sigma


import json
from aoet import logger

def parse_headers(data, is_request=False):
    headers = {}
    for h in data[1:]:
        t = h.split(b':', 1)
        headers[t[0].decode('utf-8').lower()] = t[1].decode('utf-8')
    if not is_request:
        status = data[0].split(b' ', 1)[1]
        return status.decode('utf-8'), headers
    else:
        first_line = data[0].split(b' ')
        return first_line[0].decode('utf-8'), first_line[1].decode('utf-8'), headers


def parse_body(headers, data):
    body = None
    if 'content-type' in headers:
        if 'json' in headers['content-type']:
            body = json.loads(data.decode('utf-8'))
        else:
            logger.warn('Unknown content-type: ' + headers['content-type'])
    return body


class HttpRequest:
    @staticmethod
    def from_socket(ind, data):
        return HttpRequest(ind, data)

    @staticmethod
    def from_json(j):
        return HttpRequest(j['id'], None, j['method'], j['path'], j['headers'], j['body'])

    def __init__(self, ind, data, method=None, path=None, headers=None, body=None):
        self._ind = ind
        self._data_bytes = None
        self._remaining_data = 0
        if data is None:
            self._method, self._path, self._headers, self._body = method, path, headers, body
        else:
            self._method, self._path, self._headers, self._body = self._parse_request(data)

    def to_dict(self):
        return {'id': self.ind, 'method': self._method, 'path': self._path, 'headers': self._headers, 'body': self._body}

    def __str__(self):
        return str(self.to_dict())

    def _parse_request(self, data):
        parts = data.split(b'\r\n\r\n', 1)

        method, path, headers = parse_headers(parts[0].split(b'\r\n'), is_request=True)

        body = None

        if 'content-length' in headers:
            self._remaining_data = int(headers['content-length']) - len(parts[1])
            if self._remaining_data <= 0:
                body = parse_body(headers, parts[1])
            else:
                self._data_bytes = parts[1]
        else:
            body = parse_body(headers, parts[1])
            self._remaining_data = 0

        return method, path, headers, body

    def append_data(self, data):
        self._data_bytes += data
        self._remaining_data -= len(data)
        if self._remaining_data <= 0:
            self._body = parse_body(self._headers, self._data_bytes)
            self._data_bytes = None

    @property
    def remaining_bytes(self):
        return self._remaining_data

    @property
    def ind(self):
        return self._ind

    @property
    def method(self):
        return self._method

    @property
    def path(self):
        return self._path


class HttpResponse:
    @staticmethod
    def from_socket(ind, data):
        return HttpResponse(ind, data)

    @staticmethod
    def from_json(j):
        return HttpResponse(j['id'], None, j['status'], j['headers'], j['body'])

    @staticmethod
    def timeout(id):
        return HttpResponse(id, None, 504)

    def __init__(self, ind, data, status=None, headers=None, body=None):
        self._ind = ind
        self._data_bytes = None
        self._remaining_data = 0
        if data is None:
            self._status, self._headers, self._body = status, headers, body
        else:
            self._status, self._headers, self._body = self.parse_response(data)

    def append_data(self, data):
        self._data_bytes += data
        self._remaining_data -= len(data)
        if self._remaining_data <= 0:
            self._body = parse_body(self._headers, self._data_bytes)
            self._data_bytes = None

    def to_dict(self):
        return {'id': self.ind, 'status': self.status, 'headers': self.headers, 'body': self.body}

    def __str__(self):
        return str(self.to_dict())

    def parse_response(self, data):
        parts = data.split(b'\r\n\r\n', 1)
        status, headers = parse_headers(parts[0].split(b'\r\n'))
        body = None
        if 'content-length' in headers:
            self._remaining_data = int(headers['content-length']) - len(parts[1])
            if self._remaining_data <= 0:
                body = parse_body(headers, parts[1])
            else:
                self._data_bytes = parts[1]
        else:
            body = parse_body(headers, parts[1])
            self._remaining_data = 0
        return status, headers, body

    @property
    def ind(self):
        return self._ind

    @property
    def status(self):
        return self._status

    @property
    def headers(self):
        return self._headers

    @property
    def body(self):
        return self._body

    @property
    def remaining_bytes(self):
        return self._remaining_data



from aoet import logger, BODY_MATRIX, UNIQUES_LIST
import aoet.report.generator as generator
import aoet.analysis.pca as pca

import sklearn.cluster as cluster

import pandas as pd
import numpy as np
import json
import os


def clusterize(X, epsilon, n_jobs):
    """
    Performs DBSCAN for provided matrix.
    
    :param X: Input matrix
    :param epsilon: input parameter for DBSCAN
    :param n_jobs: Number of parallel jobs to run
    :return: 
    
    Reference
    ----------
    http://scikit-learn.org/stable/auto_examples/cluster/plot_dbscan.html#sphx-glr-auto-examples-cluster-plot-dbscan-py
    """
    min_samples = 10
    logger.info('Performing DBSCAN with eps={} and min_samples={}'.format(epsilon, min_samples))

    db = cluster.DBSCAN(eps=epsilon, min_samples=min_samples, n_jobs=n_jobs).fit(X)

    core_samples_mask = np.zeros_like(db.labels_, dtype=bool)
    core_samples_mask[db.core_sample_indices_] = True
    labels = db.labels_

    # Number of clusters in labels, ignoring noise if present.
    n_clusters = len(set(labels)) - (1 if -1 in labels else 0)
    outliers = len(labels[labels == -1])

    logger.info('DBSCAN returned {} clusters and {} outliers'.format(n_clusters, outliers))

    return labels, core_samples_mask, n_clusters, outliers


def generate(arguments):
    if arguments.raw_dir is None:
        arguments.raw_dir = arguments.in_dir

    with open(os.path.join(arguments.in_dir, UNIQUES_LIST)) as fp:
        uniques = np.array(json.load(fp))

    frame = pd.read_csv(os.path.join(arguments.in_dir, BODY_MATRIX))
    if frame.shape[1] > 3:
        dim2 = pca.pca_projection(frame.values[:, 1:], 2)
    else:
        dim2 = frame.values[:, 1:]

    labels, core, n_clusters, outliers = clusterize(dim2, float(arguments.epsilon), arguments.n_jobs)

    core_samples = []

    for k in range(n_clusters):
        m = (labels == k) & core
        for i in range(len(m)):
            if m[i]:
                core_samples.append(i)
                break

    generator.to_report(dim2, arguments.raw_dir, uniques, arguments.out, labels, core_samples, n_clusters, outliers)



def add_subparsers(subparsers):
    parser_a = subparsers.add_parser('report', help='Produces final report')

    parser_a.add_argument('--in',  dest='in_dir', type=str, default='out', required=True, help='Location of converted matrix')
    parser_a.add_argument('--raw', dest='raw_dir', type=str, default=None, help='Location of raw requests. By default is'
                                                                                ' assumed to be equal to input directory')
    parser_a.add_argument('--epsilon',  dest='epsilon',  type=float, default=0.035, help='Epsilon value for DBSCAN')
    parser_a.add_argument('--out', '-o', dest='out', type=str, default='report', help='Folder for the final report')
    parser_a.add_argument('--jobs', dest='n_jobs', type=int, default=-1, help='Number of parallel jobs to run')
    parser_a.set_defaults(func=generate)


if __name__ == '__main__':
    from argparse import Namespace
    n = Namespace()
    n.in_dir = '../../../data/real1_real2'
    n.out = '../../../reports/real1_real2'

    n.raw_dir = None
    n.epsilon = 0.035
    n.n_jobs = -1

    generate(n)


import numpy as np

class CLIQUE:
    def __init__(self, X, epsilon, tau):
        """
        Initializes CLIQUE and creates density matrix
        :param X: Input Feature Matrix 
        :param epsilon: ammount of partitions
        :param tau: density threshold
        """
        self.X = X
        self.epsilon = epsilon
        self.tau = tau

        self._D = np.zeros((tau,) * X.shape[1])

        for i in range(0):
            pass



import aoet
import aoet.analysis.pca as pca
from aoet import logger, DATA_FILE

from shutil import copyfile

import pandas as pd
import numpy as np
import json
import os


def to_report(matrix, raw_data_dir, uniques, out, labels, core_samples, n_clusters, outliers):
    cols = ['x', 'y', 'label']
    if matrix.shape[1] > 2:
        try:
            matrix = pca.pca_projection(matrix)
        except np.linalg.LinAlgError as e:
            logger.exception(e)
            logger.warn(matrix)
            raise e
    elif matrix.shape[1] == 1:
        cols = ['x']
    elif matrix.shape[1] == 0:
        logger.error("No data to output!")
        return

    if not os.path.exists(out):
        os.makedirs(out)

    with open(os.path.abspath(os.path.join(raw_data_dir, aoet.REQUESTS_FILE))) as fp:
        req = np.array(json.load(fp))[uniques]
    with open(os.path.join(raw_data_dir, aoet.PRD_FILE)) as fp:
        prd = np.array(json.load(fp))[uniques]
    with open(os.path.join(raw_data_dir, aoet.TST_FILE)) as fp:
        tst = np.array(json.load(fp))[uniques]

    with open(os.path.abspath(os.path.join(out, aoet.REQUESTS_FILE)), 'w') as fp:
        json.dump(req.tolist(), fp)
    with open(os.path.join(out, aoet.PRD_FILE), 'w') as fp:
        json.dump(prd.tolist(), fp)
    with open(os.path.join(out, aoet.TST_FILE), 'w') as fp:
        json.dump(tst.tolist(), fp)
    with open(os.path.join(out, 'stats.json'), 'w') as fp:
        json.dump({'labels': labels.tolist(), 'clusters': n_clusters, 'outliers': outliers, 'core_samples': core_samples}, fp)

    frame = pd.DataFrame(np.hstack((matrix, labels.reshape(matrix.shape[0], 1))), columns=cols)

    with open(os.path.join(out, DATA_FILE), 'w') as fp:
        fp.write(frame.to_json(orient='records'))

    copyfile(os.path.abspath(os.path.join(os.path.dirname(__file__), 'resources', 'results.html')),
             os.path.join(out, 'results.html'))

    text = "Results: {} clusters and {} outliers. \n".format(n_clusters, outliers)
    for i in range(n_clusters):
        text += '-' * 60 + '\n'
        text += 'Cluster {}, most typical example: \n'.format(i)
        text += ">>> Request"
        text += json.dumps(req[core_samples[i]], indent=2, ensure_ascii=False)
        text += "\n <<< PRD \n"
        text += json.dumps(prd[core_samples[i]], indent=2, ensure_ascii=False)
        text += "\n <<< TST \n"
        text += json.dumps(tst[core_samples[i]], indent=2, ensure_ascii=False)
        text += '\n' + '-' * 60 + '\n'

    logger.info(text)
    with open(os.path.join(out, 'overview.txt'), "w") as text_file:
        text_file.write(text)


import aoet.report.generator as generator
import aoet.proxy.splitter as splitter
import aoet

import pandas as pd
import numpy as np
import difflib
import json
import os

CONVERTED_DATA_TYPE = np.float64
OBJECT_SEPARATOR = '.'
DIFFERENT_TYPE_DIFF = CONVERTED_DATA_TYPE(1)
UNDEFINED_DIFF = CONVERTED_DATA_TYPE(1)
NULL_DIFF = CONVERTED_DATA_TYPE(1)


class MatrixParser:
    def __init__(self, in_dir, out, function):
        self.in_dir = in_dir
        self.out = out
        self.function = function

    @staticmethod
    def _zip_keys(a, b):
        t = list(a.keys()) if a is not None else []
        if b is not None:
            for i in b.keys():
                if i not in t:
                    t.append(i)
        return t

    def compare_dicts(self, prefix, a, b):
        a_k = set(a.keys())
        b_k = set(b.keys())
        ret = list(map(lambda x: prefix + OBJECT_SEPARATOR + x, a_k.symmetric_difference(b_k)))
        for k in a_k.intersection(b_k):
            if a[k] != b[k]:
                if isinstance(a[k], dict) and isinstance(b[k], dict):
                    ret.extend(self.compare_dicts(prefix + OBJECT_SEPARATOR + k, a[k], b[k]))
                else:
                    ret.append(prefix + OBJECT_SEPARATOR + k)
        return set(ret)

    def first_iteration(self, prd, tst):
        diff_body = set()
        diffs = set()

        for i in range(len(prd)):
            p_i = prd[i]
            t_i = tst[i]
            for b in self._zip_keys(p_i.body, t_i.body):
                if p_i.body is None or t_i.body is None:
                    diff_body.add(b)
                    diffs.add(i)
                elif b not in p_i.body or b not in t_i.body \
                        or (p_i.body[b] is None and t_i.body[b] is not None) \
                        or (t_i.body[b] is None and p_i.body[b] is not None):
                    if b in p_i.body and p_i.body[b] is not None and type(p_i.body[b]) == dict:
                        diff_body = diff_body.union(set(map(lambda x: b + OBJECT_SEPARATOR + x, p_i.body[b].keys())))
                    elif b in t_i.body and t_i.body[b] is not None and type(t_i.body[b]) == dict:
                        diff_body = diff_body.union(set(map(lambda x: b + OBJECT_SEPARATOR + x, t_i.body[b].keys())))
                    else:
                        diff_body.add(b)
                    diffs.add(i)
                else:
                    if p_i.body[b] != t_i.body[b] or not isinstance(p_i.body[b], type(t_i.body[b])):
                        diffs.add(i)
                        if isinstance(p_i.body[b], dict):
                            if isinstance(t_i.body[b], dict):
                                diff_body = diff_body.union(self.compare_dicts(b, p_i.body[b], t_i.body[b]))
                            else:
                                diff_body = diff_body.union(
                                    set(map(lambda x: b + OBJECT_SEPARATOR + x, p_i.body[b].keys())))
                        elif isinstance(t_i.body[b], dict):
                            diff_body = diff_body.union(set(map(lambda x: b + OBJECT_SEPARATOR + x, t_i.body[b].keys())))
                        else:
                            diff_body.add(b)

        uniques = list(diffs)
        uniques.sort()
        return list(diff_body), uniques

    def cmp_numbers(self, a, b):
        if a == np.float32(0) or b == np.float32(0):
            return 0.5
        if np.abs(a) > np.abs(b):
            return np.abs(b / a)
        else:
            return np.abs(a / b)

    def to_hashable(self, a):
        if type(a) == dict:
            return ','.join(['%s:%s' % (key, value) for (key, value) in sorted(a.items())])
        elif type(a) == list:
            return ','.join(map(str, a))
        else:
            return a

    def set_seqs(self, a, b):
        t = difflib.SequenceMatcher()
        t.set_seqs(list(map(self.to_hashable, a)), list(map(self.to_hashable, b)))
        return t

    def to_value(self, a, b):
        if isinstance(a, type(b)):
            if a == b:
                return 0
            if type(a) == bool:
                return 0.5
            elif type(a) == int or type(a) == float:
                ta, tb = CONVERTED_DATA_TYPE(a), CONVERTED_DATA_TYPE(b)
                return self.function(self.cmp_numbers(ta, tb))
            elif type(a) == str:
                t = difflib.SequenceMatcher()
                t.set_seqs(a, b)
                r = t.ratio()
                return self.function(CONVERTED_DATA_TYPE(1 - r))
            elif type(a) == list or type(a) == dict:
                r = self.set_seqs(a, b).ratio()
                return self.function(CONVERTED_DATA_TYPE(1 - r))
        else:
            return DIFFERENT_TYPE_DIFF

    def compare_values(self, a, b, key):
        keys = key.split(OBJECT_SEPARATOR)
        if a is None and b is None:
            return 0
        elif a is None or b is None:
            if a is None and keys[0] in b:
                obj = b[keys[0]]
            elif b is None and keys[0] in a:
                obj = a[keys[0]]
            else:
                return 0
            for k in keys[1:]:
                if k in obj:
                    obj = obj[k]
                else:
                    obj = None
                    break
            if obj is None:
                return 0
            else:
                return UNDEFINED_DIFF

        if keys[0] not in a and keys[0] not in b:
            return 0

        if len(keys) > 1:
            if keys[0] in a and keys[0] in b:
                if isinstance(a[keys[0]], dict) and isinstance(b[keys[0]], dict):
                    return self.compare_values(a[keys[0]], b[keys[0]], OBJECT_SEPARATOR.join(keys[1:]))
                elif isinstance(a[keys[0]], dict) or isinstance(b[keys[0]], dict):
                    return DIFFERENT_TYPE_DIFF if isinstance(a[keys[0]], dict) else DIFFERENT_TYPE_DIFF
                else:
                    return 0
            elif keys[0] in a or keys[0] in b:
                return UNDEFINED_DIFF if keys[0] in a else UNDEFINED_DIFF
            else:
                return 0
        else:
            if keys[0] in a and keys[0] in b:
                return self.to_value(a[keys[0]], b[keys[0]])
            else:
                return NULL_DIFF if keys[0] in a else NULL_DIFF

    def second_iteration(self, prd, tst, diff_body, uniques=None):
        if uniques is None:
            b_matrix = np.zeros((len(prd), len(diff_body)))
            uniques = range(len(prd))
        else:
            b_matrix = np.zeros((len(uniques), len(diff_body)))

        i = 0
        for u in uniques:
            p_u = prd[u]
            t_u = tst[u]
            for j in range(len(diff_body)):
                b = diff_body[j]
                v = self.compare_values(p_u.body, t_u.body, b)
                b_matrix[i, j] = v
            i += 1

        return b_matrix

    def _parse(self):
        req, prd, tst = splitter.ResponsesContainer.load(self.in_dir)
        body, unique = self.first_iteration(prd, tst)
        body_matr = self.second_iteration(prd, tst, body, unique)

        pd.DataFrame(body_matr, columns=body).to_csv(os.path.join(self.out, aoet.BODY_MATRIX), columns=body)

        with open(os.path.join(self.out, aoet.UNIQUES_LIST), 'w') as fp:
            json.dump(unique, fp)

        return body_matr, unique


def parse(arguments):
    func = lambda x: x
    if arguments.function == 'tanh':
        func = lambda x: np.tahn(8 * x)
    mp = MatrixParser(arguments.in_dir, arguments.out, func)
    mp._parse()


def add_subparsers(subparsers):
    parser_a = subparsers.add_parser('parse', help='Converts raw input data to a matrix format, '
                                                   'filters out duplicating responses')
    parser_a.add_argument('--in', '-i', dest='in_dir', type=str, default='./out', help='Input files folder with prd, '
                                                                                       'tst and requests json files')
    parser_a.add_argument('--out', '-o', dest='out', type=str, default='./out', help='Output folder')
    parser_a.add_argument('--func', dest='function', type=str, default='lin', choices=['lin', 'tanh'],
                          help='Apply linear/tanh function to converted values. While linear function results in large '
                               'stretched clusters which are easily grouped by DBSCAN, tanh outputs more groups of '
                               'narrow clusters which might be easier to check')
    parser_a.set_defaults(func=parse)



import argparse
import logging
import sys

__version__ = '0.4'


REQUESTS_FILE = 'requests.json'
TST_FILE = 'tst.json'
PRD_FILE = 'prd.json'

MATRIX_IMAGE = 'points.png'

BODY_MATRIX  = 'body.csv'
UNIQUES_LIST = 'uniques.json'

DATA_FILE = 'results.json'

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('aoet')

args = sys.argv[1:]
if '--debug' in args:
    args.remove('--debug')
    logger.setLevel(logging.DEBUG)
else:
    logger.setLevel(logging.INFO)

if '--no-logo' in args:
    args.remove('--no-logo')
else:
    logger.info('-' * 13 + 'Automation of Exploratory Testing ' + '-' * 13)
    logger.info(' ' * 30 + 'v ' + __version__)
    logger.info('-' * 60)


import aoet.parser.matrix_parser as matrix_parser
import aoet.analysis.analyzer as analyzer
import aoet.proxy.splitter as splitter

parser = argparse.ArgumentParser(prog='aoet', description='Automation of Exploratory Testing')
parser.add_argument('--version', action='version', version=__version__)
subparsers = parser.add_subparsers()

splitter.add_subparsers(subparsers)
matrix_parser.add_subparsers(subparsers)
analyzer.add_subparsers(subparsers)

def main():
    parsed = parser.parse_args(args=args)
    if hasattr(parsed, 'func'):
        func = parsed.func
        func(parsed)



import threading
import logging
import socket
import json
from argparse import Namespace

try:
    import ssl
except Exception as e:
    print("Python failed to import ssl package, hope you won't need it")
import sys
import re
import os

import aoet.proxy.http as http
import aoet
from aoet import logger

host_re = re.compile(b'(Host:\ )(.+)$', re.MULTILINE)
location_re = re.compile(b'(Location:\ )(.+)$', re.MULTILINE)


class Splitter:
    def __init__(self, args):
        self.prd = args.prd
        self.tst = args.tst
        self.prd_port = args.prd_port
        self.tst_port = args.tst_port

        self.port = args.port
        self.container = ResponsesContainer(args.out)

        self.buffer_size = args.buffer_size

        self._should_shutdown = False

    def start(self):
        i = 0
        logger.info("Starting listening for incoming traffic on port " + str(self.port))
        logger.info("Splitting traffic between {}:{} and {}:{}".format(self.prd, self.prd_port, self.tst, self.tst_port))
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
            s.bind(('localhost', self.port))
            s.listen(0)
        except Exception as e:
            logger.exception("Error during initializing connection", e)
            sys.exit(2)

        logger.info("Server is ready to accept connections")

        self._should_shutdown = False

        try:
            while not self._should_shutdown:
                try:
                    conn, addr = s.accept()
                except socket.timeout:
                    continue

                data = conn.recv(self.buffer_size)
                full_request = data
                req = http.HttpRequest.from_socket(i, data)

                while req.remaining_bytes > 0:
                    data = conn.recv(self.buffer_size)
                    full_request += data
                    req.append_data(data)

                self.container.append_request(req)
                threading.Thread(target=self._forward,
                                 args=(i,
                                       self.prd,
                                       self.prd_port,
                                       self._replace_host(full_request, self.prd.encode('utf-8')),
                                       conn)
                                 ).start()
                threading.Thread(target=self._forward,
                                 args=(i,
                                       self.tst,
                                       self.tst_port,
                                       self._replace_host(full_request, self.tst.encode('utf-8')))
                                 ).start()
                i += 1
        except KeyboardInterrupt:
            logger.info("Keyboard interrupt received, shutting down")
            sys.exit(0)
        except Exception as e:
            logger.exception(e)
            sys.exit(2)
        finally:
            logger.info("Finally")
            s.close()
            self.container.save()

    def stop(self):
        self._should_shutdown = True

    def _forward(self, ind, dest, port, data, conn=None):
        s = None
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            logger.debug('Connecting to %s %s' % (str(dest), str(port)))
            s.connect((dest, port))
            if port == 443:
                s = ssl.wrap_socket(s, keyfile=None, certfile=None, server_side=False, cert_reqs=ssl.CERT_NONE)
            s.settimeout(5)
            s.sendall(data)
            response = None
            while response is None or response.remaining_bytes > 0:
                reply = s.recv(self.buffer_size)
                if len(reply) > 0:
                    if response is None:
                        response = http.HttpResponse.from_socket(ind, reply)
                    else:
                        response.append_data(reply)
                    if conn is not None:
                        conn.send(reply)
            s.close()

            if conn is not None:
                self.container.append_prd_response(response)
                conn.close()
            else:
                self.container.append_tst_response(response)
        except socket.timeout:
            if conn is not None:
                logging.warning("Timeout in PRD thread")
                self.container.append_prd_response(http.HttpResponse.timeout(ind))
                conn.close()
            else:
                logger.warning("Timeout in TST thread")
                self.container.append_tst_response(http.HttpResponse.timeout(ind))
        except Exception as e:
            logger.exception(e)
            if conn is not None:
                conn.close()
            s.close()
            sys.exit(1)

    def _replace_host(self, req, repl):
        global host_re
        return host_re.sub(b'Host: ' + repl + b'\r', req)

    def _replace_location(self, req, repl):
        global location_re
        return location_re.sub(b'Location: ' + repl + b'\r', req)


class ResponsesContainer:
    def __init__(self, out):
        self._prd_resp = []
        self._tst_resp = []
        self._reqs = []
        self._out = out
        if not os.path.exists(out):
            os.makedirs(out)

    def append_request(self, req):
        self._reqs.append(req)

    def append_prd_response(self, resp):
        self._prd_resp.append(resp)

    def append_tst_response(self, resp):
        self._tst_resp.append(resp)

    def save(self):
        self._reqs = list(map(http.HttpRequest.to_dict, sorted(self._reqs, key=lambda a: a.ind)))
        self._prd_resp = list(map(http.HttpResponse.to_dict, sorted(self._prd_resp, key=lambda a: a.ind)))
        self._tst_resp = list(map(http.HttpResponse.to_dict, sorted(self._tst_resp, key=lambda a: a.ind)))

        # self.validate()
        logger.info("Saving {} requests and {} prd and {} tst responses.".format(len(self._reqs), len(self._prd_resp), len(self._tst_resp)))

        with open(os.path.join(self._out, aoet.REQUESTS_FILE), 'w') as fp:
            json.dump(self._reqs, fp)
        with open(os.path.join(self._out, aoet.PRD_FILE), 'w') as fp:
            json.dump(self._prd_resp, fp)
        with open(os.path.join(self._out, aoet.TST_FILE), 'w') as fp:
            json.dump(self._tst_resp, fp)

    @staticmethod
    def load(folder):
        with open(os.path.abspath(os.path.join(folder, aoet.REQUESTS_FILE))) as fp:
            j = json.load(fp)
            requests = list(map(http.HttpRequest.from_json, j))
        with open(os.path.join(folder, aoet.PRD_FILE)) as fp:
            j = json.load(fp)
            prd = list(map(http.HttpResponse.from_json, j))
        with open(os.path.join(folder, aoet.TST_FILE)) as fp:
            j = json.load(fp)
            tst = list(map(http.HttpResponse.from_json, j))
        return requests, prd, tst


def start(arguments):
    s = Splitter(arguments)
    s.start()


def add_subparsers(subparsers):
    parser_split = subparsers.add_parser('splitter', help='Forking Traffic Splitter')

    parser_split.add_argument('--prd',      dest='prd',      type=str, required=True, help='PRD server address')
    parser_split.add_argument('--prd-port', dest='prd_port', type=int, required=True, help='PRD server port')
    parser_split.add_argument('--tst',      dest='tst',      type=str, required=True, help='TST server address')
    parser_split.add_argument('--tst-port', dest='tst_port', type=int, required=True, help='TST server port')

    parser_split.add_argument('--port', '-p', dest='port', type=int, help='Proxy listening port', default=8080)
    parser_split.add_argument('--out', '-o',  dest='out',  type=str, help='Output directory', default='./out')

    parser_split.add_argument('--buffer-size', dest='buffer_size', type=int, default=4096, help='Buffer size')
    parser_split.set_defaults(func=start)


if __name__ == '__main__':
    n = Namespace()
    n.port = 8080
    n.buffer_size = 4096
    n.prd = 'acer_filestorage'
    n.prd_port = 8080
    n.tst = 'acer_filestorage'
    n.tst_port = 8081
    n.out = 'data/tmp'
    print("Starting with default config")
    start(n)



from setuptools import setup
import aoet


def readme():
    with open('README.rst') as f:
        return f.read()


setup(name='aoet',
      version=aoet.__version__,
      description='Automation of Exploratory Testing',
      ong_description=readme(),
      url='https://gitlab.tubit.tu-berlin.de/abar/aoet',
      author='Anton Bardishev',
      author_email='abar@protonmail.ch',
      license='MIT',
      packages=['aoet', 'aoet.analysis', 'aoet.parser', 'aoet.proxy', 'aoet.report'],
      install_requires=['numpy',
                        'matplotlib',
                        'sklearn',
                        'pandas'
                        ],
      scripts=['bin/aoet'],
      zip_safe=False)